import re

from typing import Any, Dict
from aws_cdk.core import RemovalPolicy
from aws_cdk.custom_resources import AwsCustomResource
from aws_empty_bucket.empty_s3_bucket import EmptyS3Bucket
from aws_empty_ecr_repository.empty_ecr_repository import EmptyEcrRepository
from aws_ci_cd_fargate.source.pipeline_commit_to_ecr import PipelineCommitToEcr
from aws_ci_cd_fargate.source.pipeline_ecr_to_ecs import PipelineEcrToEcs
from aws_cdk import (
    aws_ecs,
    aws_codecommit,
    aws_elasticloadbalancingv2,
    aws_s3,
    core
)


class EcsPipeline:
    """
    Class which creates infrastructure for CI/CD Blue/Green Ecs Fargate deployments.
    """
    def __init__(
            self,
            scope: core.Stack,
            prefix: str,
            main_listener: aws_elasticloadbalancingv2.CfnListener,
            deployments_listener: aws_elasticloadbalancingv2.CfnListener,
            ecs_service: AwsCustomResource,
            ecs_cluster: aws_ecs.Cluster,
            task_def: str,
            app_spec: str,
            build_environment: Dict[str, Any],
            docker_build_args: Dict[str, str],
            production_target_group,
            deployment_target_group
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation template to which add resources.
        :param prefix: A prefix for newly created resources.
        :param main_listener: A listener which receives incoming traffic and forwards it to a target group.
        :param deployments_listener: A listener which receives incoming traffic and forwards it to a target group.
        This listener is used for blue/green deployment.
        :param ecs_service: Ecs service to which create this pipeline.
        :param ecs_cluster: ECS cluster in which the ECS service is.
        :param task_def: Task definition object defining the parameters for a newly deployed container.
        :param app_spec: App specification object defining the ecs service modifications.
        :param build_environment: Environment variables for a build step. You can put here various config
        parameters, urls, secrets, etc.
        :param docker_build_args: Build arguments for docker build command.
        :param production_target_group: A target group where your blue instances are serving production traffic.
        :param deployment_target_group: A target group where your green instances are ready to serve production traffic.
        """
        self.artifacts_bucket = EmptyS3Bucket(
            scope,
            self.__convert(prefix + 'FargateArtifacts'),
            access_control=aws_s3.BucketAccessControl.PRIVATE,
            bucket_name=self.__convert(prefix + 'FargateArtifacts'),
        )

        self.source_code_repository = aws_codecommit.Repository(
            scope,
            prefix + 'FargateSourceCode',
            repository_name=prefix + 'FargateSourceCode'
        )

        self.ecr_repository = EmptyEcrRepository(
            scope, prefix + 'FargateEcrRepository',
            repository_name=prefix.lower(),
            removal_policy=RemovalPolicy.DESTROY
        )

        self.ecr_to_ecs = PipelineEcrToEcs(
            scope=scope,
            prefix=prefix,
            artifacts_bucket=self.artifacts_bucket,
            source_repository=self.source_code_repository,
            ecr_repository=self.ecr_repository,
            task_def=task_def,
            app_spec=app_spec,
            main_listener=main_listener,
            deployments_listener=deployments_listener,
            ecs_cluster=ecs_cluster,
            ecs_service=ecs_service,
            production_target_group=production_target_group,
            deployment_target_group=deployment_target_group
        )

        self.commit_to_ecr = PipelineCommitToEcr(
            scope=scope,
            prefix=prefix,
            artifacts_bucket=self.artifacts_bucket,
            ecr_repository=self.ecr_repository,
            source_repository=self.source_code_repository,
            build_environment=build_environment,
            docker_build_args=docker_build_args,
            next_pipeline=self.ecr_to_ecs.ecr_to_ecs_pipeline
        )

    @staticmethod
    def __convert(name: str) -> str:
        """
        Converts CamelCase string to pascal-case where underscores are dashes.
        This is required due to S3 not supporting capital letters or underscores.
        """
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()



from aws_cdk import aws_ec2
from aws_ci_cd_fargate.parameters.lb_listener_parameters import LbListenerParameters
from aws_ci_cd_fargate.parameters.pipeline_parameters import PipelineParams
from aws_ci_cd_fargate.source.ecs_main import Ecs
from aws_ci_cd_fargate.source.ecs_pipeline import EcsPipeline
from aws_ci_cd_fargate.parameters.ecs_parameters import EcsParams
from aws_ci_cd_fargate.parameters.load_balancer_parameters import LoadBalancerParams
from aws_ci_cd_fargate.source.lb_listener_config import LbListenerConfig


class EcsFargateWithCiCd:
    """
    Creates a whole infrastructure around ECS Fargate service and blue/green CI/CD deployments.
    """
    def __init__(
            self,
            scope,
            prefix: str,
            vpc: aws_ec2.Vpc,
            lb_params: LoadBalancerParams,
            ecs_params: EcsParams,
            lb_listener_params: LbListenerParameters,
            pipeline_params: PipelineParams
    ) -> None:
        """
        Constructor.

        :param scope: A CF stack in which to create resources.
        :param prefix: The prefix for all newly created resources. E.g. Wordpress.
        :param vpc: Virtual private cloud (VPC).
        :param lb_params: Loadbalancer parameters.
        :param ecs_params: Compute power parameters for newly deployed container.
        :param lb_listener_params: Parameters two configure existing listeners with listener rules.
        :param pipeline_params: Configuration parameters for ci/cd pipeline.
        """
        self.lb_listener_config = LbListenerConfig(
            scope,
            prefix=prefix,
            vpc=vpc,
            listener_params=lb_listener_params,
            healthy_http_codes=lb_params.healthy_http_codes,
            health_check_path=lb_params.health_check_path
        )

        self.ecs = Ecs(
            scope,
            prefix=prefix,
            ecs_params=ecs_params,
            lb_listener_config=self.lb_listener_config,
            vpc=vpc
        )

        self.pipeline = EcsPipeline(
            scope,
            prefix=prefix,
            main_listener=lb_listener_params.production_listener,
            deployments_listener=lb_listener_params.deployment_listener,
            ecs_service=self.ecs.service,
            ecs_cluster=self.ecs.cluster,
            task_def=self.ecs.create_task_def(),
            app_spec=self.ecs.create_appspec(),
            build_environment=pipeline_params.build_environment,
            docker_build_args=pipeline_params.docker_build_args,
            production_target_group=self.lb_listener_config.production_target_group,
            deployment_target_group=self.lb_listener_config.deployment_target_group
        )



from aws_cdk import aws_elasticloadbalancingv2


class LbListenerParameters:
    def __init__(
            self,
            production_listener: aws_elasticloadbalancingv2.CfnListener,
            deployment_listener: aws_elasticloadbalancingv2.CfnListener,
            rule_condition: aws_elasticloadbalancingv2.CfnListenerRule.RuleConditionProperty,
            rule_priority: int,
    ) -> None:
        """
        Constructor.

        :param production_listener: A loadbalancer's main (blue) listener instance.
        :param deployment_listener: A loadbalancer's test (green) traffic listener instance.
        :param rule_condition: A listener routing rule condition. When a rule is matched, the
        traffic is picked up by the listeners and sent to target groups. Read more:
        https://docs.aws.amazon.com/elasticloadbalancing/latest/application/listener-update-rules.html
        :param rule_priority: A priority for the rule. The lower the number, the higher the priority.
        Read more: https://stackoverflow.com/questions/43385942/aws-alb-rule-priority

        :return: No return.
        """
        self.production_listener = production_listener
        self.deployment_listener = deployment_listener
        self.rule_condition = rule_condition
        self.rule_priority = rule_priority



from typing import Optional, Dict, Any


class PipelineParams:
    def __init__(
            self,
            build_environment: Optional[Dict[str, Any]] = None,
            docker_build_args: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Constructor.

        :param build_environment: Environment variables for a build step. You can put here various config
        parameters, urls, secrets, etc.
        :param docker_build_args: Build arguments for docker build command.
        """
        self.build_environment: Dict[str, Any] = build_environment or {}
        self.docker_build_args: Dict[str, str] = docker_build_args or {}



import json

from aws_cdk import aws_logs, aws_ecs, aws_applicationautoscaling, aws_ec2, aws_iam
from aws_cdk.core import Stack, RemovalPolicy
from aws_ci_cd_fargate.parameters.ecs_parameters import EcsParams
from aws_ci_cd_fargate.source.custom.ecs_service import EcsService
from aws_ci_cd_fargate.source.lb_listener_config import LbListenerConfig
from aws_ecs_cluster.ecs_cluster import EcsCluster

class Ecs:
    """
    Class that creates ECS Fargate service.
    """
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            ecs_params: EcsParams,
            lb_listener_config: LbListenerConfig,
            vpc: aws_ec2.Vpc
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation template to which add resources.
        :param prefix: A prefix for newly created resources.
        :param ecs_params: Compute power parameters for newly deployed container.
        :param lb_listener_config: Listeners configuration for blue-green deployments.
        :param vpc: Virtual Private Cloud in which loadbalancer and other instances are/will be located.
        """
        self.prefix = prefix
        self.aws_region = scope.region
        self.ecs_params = ecs_params

        self.task_execution_role = aws_iam.Role(
            scope, prefix + 'FargateTaskExecutionRole',
            path='/',
            inline_policies={
                prefix + 'FargateTaskExecutionPolicy': aws_iam.PolicyDocument(
                    statements=[aws_iam.PolicyStatement(
                        actions=[
                            "ecr:GetAuthorizationToken",
                            "ecr:BatchCheckLayerAvailability",
                            "ecr:GetDownloadUrlForLayer",
                            "ecr:BatchGetImage",
                            "logs:CreateLogStream",
                            "logs:PutLogEvents",
                            "cloudtrail:LookupEvents"
                        ],
                        resources=['*'],
                        effect=aws_iam.Effect.ALLOW)]
                )},
            assumed_by=aws_iam.ServicePrincipal('ecs-tasks.amazonaws.com'),
        )

        self.log_group = aws_logs.LogGroup(
            scope, prefix + 'FargateEcsLogGroup',
            log_group_name=f'/aws/ecs/fargate/{prefix}',
            removal_policy=RemovalPolicy.DESTROY
        )

        self.cluster = EcsCluster(
            scope, prefix + 'FargateCluster',
            cluster_name=prefix + 'FargateCluster',
            vpc=vpc
        )

        self.task = aws_ecs.FargateTaskDefinition(
            scope, prefix + 'FargateTaskDefinition',
            cpu=int(self.ecs_params.container_cpu), memory_limit_mib=int(self.ecs_params.container_ram), family=prefix.lower(),
            execution_role=self.task_execution_role
        )

        self.container = self.task.add_container(
            self.ecs_params.container_name,
            image=aws_ecs.ContainerImage.from_registry('eexit/mirror-http-server:latest'),
            logging=aws_ecs.AwsLogDriver(stream_prefix=prefix, log_group=self.log_group)
        )
        self.container.add_port_mappings(aws_ecs.PortMapping(container_port=80))

        self.service = EcsService(
            stack=scope,
            prefix=prefix,
            cluster=self.cluster,
            task=self.task,
            ecs_params=self.ecs_params,
            production_target_group=lb_listener_config.production_target_group
        ).get_resource().custom_resource

        self.service.node.add_dependency(lb_listener_config.production_target_group)
        self.service.node.add_dependency(lb_listener_config.deployment_target_group)

        self.scalable_target = aws_applicationautoscaling.ScalableTarget(
            scope, prefix + 'FargateScalableTarget',
            min_capacity=1,
            max_capacity=5,
            service_namespace=aws_applicationautoscaling.ServiceNamespace.ECS,
            resource_id='/'.join(['service', self.cluster.cluster_name, prefix + 'FargateService']),
            scalable_dimension='ecs:service:DesiredCount'
        )

        self.scalable_target.node.add_dependency(self.service)

        self.scaling_policy = aws_applicationautoscaling.TargetTrackingScalingPolicy(
            scope, prefix + 'FargateScalingPolicy',
            policy_name=prefix + 'FargateScalingPolicy',
            scaling_target=self.scalable_target,
            target_value=self.ecs_params.cpu_threshold,
            predefined_metric=aws_applicationautoscaling.PredefinedMetric.ECS_SERVICE_AVERAGE_CPU_UTILIZATION,
            disable_scale_in=False
        )

    def create_appspec(self) -> str:
        """
        Creates an application specification object which will be used for deploying new containers through a pipeline.
        The application specification object specifies parameters about the ECS service.

        :return: Application specification object.
        """
        app_spec = (
            f'version: 0.0',
            f'Resources:',
            f'  - TargetService:',
            f'      Type: AWS::ECS::Service',
            f'      Properties:',
            f'        TaskDefinition: <TASK_DEFINITION>',
            f'        LoadBalancerInfo:',
            f'          ContainerName: "{self.container.container_name}"',
            f'          ContainerPort: 80',
        )

        return '\n'.join(app_spec)

    def create_task_def(self) -> str:
        """
        Creates a task definition object which will be used for deploying new containers through a pipeline.
        The task definition object specifies parameters about newly created containers.

        :return: Task definition object.
        """
        definition = {
            'executionRoleArn': self.task.execution_role.role_arn,
            'containerDefinitions': [
                {
                    'name': self.ecs_params.container_name,
                    'image': "<IMAGE1_NAME>",
                    'essential': True,
                    'environment': [
                        {'name': key, 'value': value} for key, value in self.ecs_params.container_environment.items()
                    ],
                    'portMappings': [
                        {
                            'containerPort': 80
                        }
                    ],
                    'logConfiguration': {
                        'logDriver': 'awslogs',
                        'options': {
                            'awslogs-group': self.log_group.log_group_name,
                            'awslogs-region': self.aws_region,
                            'awslogs-stream-prefix': self.prefix
                        }
                    }
                }
            ],
            'requiresCompatibilities': [
                'FARGATE'
            ],
            'networkMode': 'awsvpc',
            'cpu': str(self.ecs_params.container_cpu),
            'memory': str(self.ecs_params.container_ram),
            'family': self.prefix.lower()
        }

        return json.dumps(definition, indent=4)



from typing import Any, Dict, Optional
from aws_cdk import core, aws_ecs
from aws_cdk.aws_elasticloadbalancingv2 import CfnTargetGroup
from aws_ci_cd_fargate.parameters.ecs_parameters import EcsParams
from aws_ecs_service.ecs_service import EcsService as EcsServiceCustomResource


class EcsService:
    """
    Custom CloudFormation resource which creates a git commit action to set deployment configuration for ecs project.
    """

    def __init__(
            self,
            stack: core.Stack,
            prefix: str,
            cluster: aws_ecs.Cluster,
            task: aws_ecs.FargateTaskDefinition,
            ecs_params: EcsParams,
            production_target_group: CfnTargetGroup
    ) -> None:
        """
        Constructor.

        :param stack: A CloudFormation stack to which add this resource.
        :param prefix: Prefix for resource names.
        :param code_repository: A codecommit git repository to push configuration files for ecs deployment.
        :param task_definition: A document which describes how ecs deployment should behave.
        :param app_spec: A document which describes how ecs deployment should behave.
        """
        self.__stack = stack
        self.__prefix = prefix
        self.__cluster = cluster
        self.__task = task
        self.__ecs_params = ecs_params
        self.__production_target_group = production_target_group

    def get_resource(self) -> EcsServiceCustomResource:
        """
        Creates a custom resource to manage an ecs deployment configuration.

        :param scope: A scope in which this resource should be created.

        :return: Custom resource to manage an ecs deployment configuration.
        """
        return EcsServiceCustomResource(
            scope=self.__stack,
            id=self.__prefix + "FargateService",
            on_create_action=self.__on_create(),
            on_update_action=self.__on_update(),
            on_delete_action=self.__on_delete()
        )

    def __on_create(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_create" command.

        :return: A dictionary command.
        """
        return {
            'cluster': self.__cluster.cluster_arn,
            'serviceName': self.__prefix + 'FargateService',
            'taskDefinition': self.__task.task_definition_arn,
            'loadBalancers': [
                {
                    'containerName': self.__ecs_params.container_name,
                    'containerPort': 80,
                    'targetGroupArn': self.__production_target_group.ref
                }
            ],
            'desiredCount': 1,
            'networkConfiguration': {
                'awsvpcConfiguration': {
                    'assignPublicIp': 'DISABLED',
                    'securityGroups': [sub.security_group_id for sub in self.__ecs_params.ecs_security_groups],
                    'subnets': [sub.subnet_id for sub in self.__ecs_params.ecs_subnets],
                }
            },
            'deploymentController': {
                'type': 'CODE_DEPLOY'
            },
            'launchType': 'FARGATE'
        }

    def __on_update(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_update" command".

        :return: A dictionary command.
        """
        return {
            'cluster': self.__cluster.cluster_arn,
            'service': self.__prefix + 'FargateService',
            'healthCheckGracePeriodSeconds': 0
        }

    def __on_delete(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_delete" command".

        :return: A dictionary command.
        """
        return {
            'cluster': self.__cluster.cluster_arn,
            'service': self.__prefix + 'FargateService',
            'force': True
        }



from typing import List, Optional
from aws_cdk import aws_ec2


class LoadBalancerParams:
    def __init__(
            self,
            healthy_http_codes: Optional[List[int]] = None,
            health_check_path: Optional[str] = None
    ) -> None:
        """
        Constructor.

        :param healthy_http_codes: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a list of http codes that your service can return and should be treated as healthy.
        :param health_check_path: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a path for that ping.

        :return: No return.
        """
        self.healthy_http_codes: [List[int]] = healthy_http_codes or [200]
        self.health_check_path: str = health_check_path or '/'



from typing import Any, Dict, Optional
from aws_cdk import core
from aws_cdk.aws_codecommit import Repository
from aws_cdk.aws_codedeploy import EcsApplication
from aws_cdk.aws_ecs import Cluster
from aws_cdk.aws_elasticloadbalancingv2 import CfnListener
from aws_cdk.aws_iam import Role, PolicyStatement, PolicyDocument, Effect, ServicePrincipal, CompositePrincipal
from aws_cdk.custom_resources import AwsCustomResource, PhysicalResourceId


class DeploymentGroup:
    """
    Custom CloudFormation resource which manages a deployment group.
    """
    def __init__(
            self,
            stack: core.Stack,
            prefix: str,
            code_repository: Repository,
            task_definition: str,
            app_spec: str,
            ecs_application: EcsApplication,
            main_listener: CfnListener,
            deployments_listener: CfnListener,
            production_target_group,
            deployment_target_group,
            ecs_cluster: Cluster,
    ) -> None:
        """
        Constructor.

        :param stack: A CloudFormation stack to which add this resource.
        :param prefix: Prefix for resource names.
        :param code_repository: A codecommit git repository to push configuration files for ecs deployment.
        :param task_definition: A document which describes how ecs deployment should behave.
        :param app_spec: A document which describes how ecs deployment should behave.
        :param ecs_application: An ecs application for which the deployments are being made.
        :param main_listener: A loadbalancer's main listener for main traffic.
        :param deployments_listener: A loadbalancer's testing listener for testing traffic.
        :param ecs_cluster: An ecs cluster in which our ecs application is located.
        """
        self.__stack = stack
        self.__prefix = prefix
        self.__code_repository = code_repository
        self.__task_definition = task_definition
        self.__app_spec = app_spec
        self.__ecs_application = ecs_application
        self.__main_listener = main_listener
        self.__deployments_listener = deployments_listener
        self.__production_target_group = production_target_group
        self.__deployment_target_group = deployment_target_group
        self.__ecs_cluster = ecs_cluster

        self.__custom_resource_role = Role(
            self.__stack,
            self.__prefix + 'CustomFargateDeploymentGroupRole',
            inline_policies={
                self.__prefix + 'CustomFargateDeploymentGroupPolicy': PolicyDocument(
                    statements=[
                        PolicyStatement(
                            actions=[
                                "codedeploy:GetDeploymentGroup",
                                "codedeploy:CreateDeploymentGroup",
                                "codedeploy:DeleteDeploymentGroup",
                                "codedeploy:UpdateDeploymentGroup",
                            ],
                            resources=['*'],
                            effect=Effect.ALLOW
                        ),
                        PolicyStatement(
                            actions=[
                                "logs:CreateLogGroup",
                                "logs:CreateLogStream",
                                "logs:PutLogEvents"
                            ],
                            resources=['*'],
                            effect=Effect.ALLOW
                        )
                    ]
                )},
            assumed_by=ServicePrincipal('lambda.amazonaws.com')
        )

        self.__deployment_group_role = Role(
            self.__stack,
            self.__prefix + 'FargateDeploymentGroupRole',
            path='/',
            inline_policies={
                self.__prefix + 'FargateDeploymentGroupPolicy': PolicyDocument(
                    statements=[
                        PolicyStatement(
                            actions=[
                                "ecs:DescribeServices",
                                "ecs:CreateTaskSet",
                                "ecs:UpdateServicePrimaryTaskSet",
                                "ecs:DeleteTaskSet",
                                "elasticloadbalancing:DescribeTargetGroups",
                                "elasticloadbalancing:DescribeListeners",
                                "elasticloadbalancing:ModifyListener",
                                "elasticloadbalancing:DescribeRules",
                                "elasticloadbalancing:ModifyRule",
                                "lambda:InvokeFunction",
                                "cloudwatch:DescribeAlarms",
                                "sns:Publish",
                                "s3:GetObject",
                                "s3:GetObjectMetadata",
                                "s3:GetObjectVersion",
                                "iam:PassRole"
                            ],
                            resources=['*'],
                            effect=Effect.ALLOW
                        )
                    ]
                )
            },
            assumed_by=CompositePrincipal(
                ServicePrincipal('ecs-tasks.amazonaws.com'),
                ServicePrincipal('codedeploy.amazonaws.com')
            )
        )

    def get_resource(self):
        """
        Creates a custom resource to manage a deployment group.

        :return: Custom resource to manage a deployment group.
        """
        return AwsCustomResource(
            scope=self.__stack,
            id=self.__prefix + "CustomFargateDeploymentGroupResource",
            on_create=self.__on_create(),
            on_update=self.__on_update(),
            on_delete=self.__on_delete(),
            role=self.__custom_resource_role
        )

    @staticmethod
    def service_name() -> str:
        """
        Returns a service name that this custom resource manages.

        :return: Service name.
        """
        return 'CodeDeploy'

    def __on_create(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_create" command.

        :return: A dictionary command.
        """
        return {
            "service": self.service_name(),
            "action": "createDeploymentGroup",
            "parameters": {
                'applicationName': self.__ecs_application.application_name,
                'deploymentGroupName': self.__prefix + 'FargateDeploymentGroup',
                'deploymentConfigName': 'CodeDeployDefault.ECSAllAtOnce',
                'serviceRoleArn': self.__deployment_group_role.role_arn,
                'autoRollbackConfiguration': {
                    'enabled': True,
                    'events': ['DEPLOYMENT_FAILURE', 'DEPLOYMENT_STOP_ON_ALARM', 'DEPLOYMENT_STOP_ON_REQUEST']
                },
                'deploymentStyle': {
                    'deploymentType': 'BLUE_GREEN',
                    'deploymentOption': 'WITH_TRAFFIC_CONTROL'
                },
                'blueGreenDeploymentConfiguration': {
                    'terminateBlueInstancesOnDeploymentSuccess': {
                        'action': 'TERMINATE',
                        'terminationWaitTimeInMinutes': 5
                    },
                    'deploymentReadyOption': {
                        'actionOnTimeout': 'CONTINUE_DEPLOYMENT',
                    },
                },
                'loadBalancerInfo': {
                    'targetGroupPairInfoList': [
                        {
                            'targetGroups': [
                                {
                                    'name': self.__production_target_group.attr_target_group_name,
                                },
                                {
                                    'name': self.__deployment_target_group.attr_target_group_name,
                                },
                            ],
                            'prodTrafficRoute': {
                                'listenerArns': [
                                    self.__main_listener.ref
                                ]
                            },
                            'testTrafficRoute': {
                                'listenerArns': [
                                    self.__deployments_listener.ref
                                ]
                            }
                        },
                    ]
                },
                'ecsServices': [
                    {
                        'serviceName': self.__prefix + 'FargateService',
                        'clusterName': self.__ecs_cluster.cluster_name
                    },
                ],
            },
            "physical_resource_id": PhysicalResourceId.of(self.__prefix + 'DeploymentGroup'),
        }

    def __on_update(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_update" command".

        :return: A dictionary command.
        """
        return {
            "service": self.service_name(),
            "action": "updateDeploymentGroup",
            "parameters": {
                'applicationName': self.__ecs_application.application_name,
                'currentDeploymentGroupName': self.__prefix + 'FargateDeploymentGroup',
                'deploymentConfigName': 'CodeDeployDefault.ECSAllAtOnce',
                'serviceRoleArn': self.__deployment_group_role.role_arn,
                'autoRollbackConfiguration': {
                    'enabled': True,
                    'events': ['DEPLOYMENT_FAILURE', 'DEPLOYMENT_STOP_ON_ALARM', 'DEPLOYMENT_STOP_ON_REQUEST']
                },
                'deploymentStyle': {
                    'deploymentType': 'BLUE_GREEN',
                    'deploymentOption': 'WITH_TRAFFIC_CONTROL'
                },
                'blueGreenDeploymentConfiguration': {
                    'terminateBlueInstancesOnDeploymentSuccess': {
                        'action': 'TERMINATE',
                        'terminationWaitTimeInMinutes': 5
                    },
                    'deploymentReadyOption': {
                        'actionOnTimeout': 'CONTINUE_DEPLOYMENT',
                    },
                },
                'loadBalancerInfo': {
                    'targetGroupPairInfoList': [
                        {
                            'targetGroups': [
                                {
                                    'name': self.__production_target_group.attr_target_group_name,
                                },
                                {
                                    'name': self.__deployment_target_group.attr_target_group_name,
                                },
                            ],
                            'prodTrafficRoute': {
                                'listenerArns': [
                                    self.__main_listener.ref
                                ]
                            },
                            'testTrafficRoute': {
                                'listenerArns': [
                                    self.__deployments_listener.ref
                                ]
                            }
                        },
                    ]
                },
                'ecsServices': [
                    {
                        'serviceName': self.__prefix + 'FargateService',
                        'clusterName': self.__ecs_cluster.cluster_name
                    },
                ],
            },
            "physical_resource_id": PhysicalResourceId.of(self.__prefix + 'DeploymentGroup'),
        }

    def __on_delete(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_delete" command".

        :return: A dictionary command.
        """


        return {
            "service": self.service_name(),
            "action": "deleteDeploymentGroup",
            "parameters": {
                'applicationName': self.__ecs_application.application_name,
                'deploymentGroupName': self.__prefix + 'FargateDeploymentGroup',
            },
            "physical_resource_id": PhysicalResourceId.of(self.__prefix + 'DeploymentGroup'),
        }



from typing import List, Optional
from aws_cdk import aws_elasticloadbalancingv2
from aws_cdk.aws_ec2 import Vpc
from aws_ci_cd_fargate.parameters.lb_listener_parameters import LbListenerParameters


class LbListenerConfig:
    """
    Class which configures loabalancer's listeners to support blue-green deployments.
    """
    TARGET_GROUP_PORT = 80

    def __init__(
            self,
            scope,
            prefix: str,
            vpc: Vpc,
            listener_params: LbListenerParameters,
            healthy_http_codes: Optional[List[int]] = None,
            health_check_path: Optional[str] = None
    ) -> None:
        """
        Constructor.

        :param scope: A CloudFormation template to which add resources.
        :param prefix: A prefix for resource names.
        :param vpc: Virtual private cloud in which target groups and a loadbalancer exist.
        :param listener_params: Configuration parameters for loadbalancer's listeners.
        :param healthy_http_codes: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a list of http codes that your service can return and should be treated as healthy.
        :param health_check_path: The deployed instance is constantly pinged to determine if it is available
        (healthy) or not. Specify a path to ping.
        """
        # By default a healthy http code is considered to be 200.
        healthy_http_codes = [str(code) for code in healthy_http_codes] if healthy_http_codes else ['200']
        healthy_http_codes = ','.join(healthy_http_codes)

        # If your service's task definition uses the awsvpc network mode
        # (which is required for the Fargate launch type), you must choose ip as the target type,
        # not instance, when creating your target groups because
        # tasks that use the awsvpc network mode are associated with an elastic network interface,
        # not an Amazon EC2 instance.
        # self.target_type = aws_elasticloadbalancingv2.TargetType.IP
        self.target_type = 'ip'

        """
        PRODUCTION CONFIG
        """

        self.production_target_group = aws_elasticloadbalancingv2.CfnTargetGroup(
            scope,
            prefix + 'FargateProdTG',
            name=prefix + 'FargateProdTG',
            matcher=aws_elasticloadbalancingv2.CfnTargetGroup.MatcherProperty(http_code=healthy_http_codes),
            port=self.TARGET_GROUP_PORT,
            protocol='HTTP',
            vpc_id=vpc.vpc_id,
            target_type=self.target_type,
            health_check_path=health_check_path if health_check_path else '/'
        )

        self.production_rule = aws_elasticloadbalancingv2.CfnListenerRule(
            scope=scope,
            id=prefix + 'ProductionListenerRule',
            actions=[
                aws_elasticloadbalancingv2.CfnListenerRule.ActionProperty(
                    type='forward',
                    target_group_arn=self.production_target_group.ref
                )
            ],
            conditions=[
                listener_params.rule_condition
            ],
            listener_arn=listener_params.production_listener.ref,
            priority=listener_params.rule_priority
        )

        """
        DEPLOYMENT CONFIG
        """

        self.deployment_target_group = aws_elasticloadbalancingv2.CfnTargetGroup(
            scope, prefix + 'FargateDeplTG',
            name=prefix + 'FargateDeplTG',
            matcher=aws_elasticloadbalancingv2.CfnTargetGroup.MatcherProperty(http_code=healthy_http_codes),
            port=self.TARGET_GROUP_PORT,
            protocol='HTTP',
            vpc_id=vpc.vpc_id,
            target_type=self.target_type,
            health_check_path=health_check_path if health_check_path else '/'
        )

        self.deployment_rule = aws_elasticloadbalancingv2.CfnListenerRule(
            scope=scope,
            id=prefix + 'DeploymentListenerRule',
            actions=[
                aws_elasticloadbalancingv2.CfnListenerRule.ActionProperty(
                    type='forward',
                    target_group_arn=self.deployment_target_group.ref
                )
            ],
            conditions=[
                listener_params.rule_condition
            ],
            listener_arn=listener_params.deployment_listener.ref,
            priority=listener_params.rule_priority
        )



import copy

from typing import Dict, Any
from aws_cdk import aws_codepipeline, aws_codepipeline_actions, aws_codecommit, aws_iam, aws_codebuild, aws_ecr
from aws_cdk.aws_codepipeline import IPipeline
from aws_cdk.aws_s3 import IBucket
from aws_cdk.core import Stack


class PipelineCommitToEcr:
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            artifacts_bucket: IBucket,
            ecr_repository: aws_ecr.Repository,
            source_repository: aws_codecommit.Repository,
            build_environment: Dict[str, Any],
            docker_build_args: Dict[str, str],
            next_pipeline: IPipeline
    ):
        self.region = scope.region
        self.ecr_repository = ecr_repository
        self.build_environment = build_environment
        self.next_pipeline = next_pipeline

        self.source_artifact = aws_codepipeline.Artifact(
            artifact_name=prefix + 'FargateCodeCommitSourceArtifact',
        )

        self.source_action = aws_codepipeline_actions.CodeCommitSourceAction(
            repository=source_repository,
            branch='master',
            action_name='CodeCommitSource',
            run_order=1,
            trigger=aws_codepipeline_actions.CodeCommitTrigger.EVENTS,
            output=self.source_artifact
        )

        docker_build_command = 'docker build -t $REPOSITORY_URI:latest .'

        for key, value in docker_build_args.items():
            docker_build_command += f' --build-arg {key}={value}'

        self.docker_build = aws_codebuild.PipelineProject(
            scope, prefix + 'FargateCodeBuildProject',
            project_name=prefix + 'FargateCodeBuildProject',
            environment_variables=self.build_environment_variables(),
            environment=aws_codebuild.BuildEnvironment(
                build_image=aws_codebuild.LinuxBuildImage.UBUNTU_14_04_DOCKER_18_09_0,
                compute_type=aws_codebuild.ComputeType.SMALL,
                privileged=True
            ),
            build_spec=aws_codebuild.BuildSpec.from_object(
                {
                    'version': 0.2,
                    'phases': {
                        'pre_build': {
                            'commands': f'$(aws ecr get-login --no-include-email --region $REGION)'
                        },
                        'build': {
                            'commands': docker_build_command,
                        },
                        'post_build': {
                            'commands': [
                                'docker push $REPOSITORY_URI:latest',
                                f'aws codepipeline start-pipeline-execution --name $PIPELINE_NAME'
                            ]
                        },
                    }
                }
            ),

        )

        self.docker_build.role.add_to_policy(
            statement=aws_iam.PolicyStatement(
                actions=[
                    "ecr:CompleteLayerUpload",
                    "ecr:GetAuthorizationToken",
                    "ecr:UploadLayerPart",
                    "ecr:InitiateLayerUpload",
                    "ecr:BatchCheckLayerAvailability",
                    "ecr:PutImage",
                    "codepipeline:StartPipelineExecution"
                ],
                resources=['*'],
                effect=aws_iam.Effect.ALLOW)
        )

        self.codecommit_to_ecr_pipeline = aws_codepipeline.Pipeline(
            scope,
            prefix + 'FargateCodeCommitToEcrPipeline',
            pipeline_name=prefix + 'FargateCodeCommitToEcrPipeline',
            artifact_bucket=artifacts_bucket,
            stages=[
                aws_codepipeline.StageProps(
                    stage_name='SourceStage',
                    actions=[self.source_action]
                ),
                aws_codepipeline.StageProps(
                    stage_name='BuildStage',
                    actions=[
                        aws_codepipeline_actions.CodeBuildAction(
                            input=self.source_artifact,
                            project=self.docker_build,
                            action_name='BuildAction',
                            run_order=1
                        )
                    ]
                )
            ]
        )

    def build_environment_variables(self):
        base_environment = {
            'REPOSITORY_URI': aws_codebuild.BuildEnvironmentVariable(value=self.ecr_repository.repository_uri),
            'PIPELINE_NAME': aws_codebuild.BuildEnvironmentVariable(value=self.next_pipeline.pipeline_name),
            'REGION': aws_codebuild.BuildEnvironmentVariable(value=self.region)
        }

        build_environment = copy.deepcopy(self.build_environment)
        build_environment.pop('REPOSITORY_URI', None)
        build_environment.pop('PIPELINE_NAME', None)
        build_environment.pop('REGION', None)

        for key, value in build_environment.items():
            if not isinstance(value, aws_codebuild.BuildEnvironmentVariable):
                build_environment[key] = aws_codebuild.BuildEnvironmentVariable(value=value)

        return {**base_environment, **build_environment}



from typing import Any, Dict, Optional
from aws_cdk import core
from aws_cdk.aws_codecommit import Repository
from aws_cdk.aws_iam import Role, PolicyStatement, PolicyDocument, Effect, ServicePrincipal
from aws_cdk.custom_resources import AwsCustomResource, PhysicalResourceId


class DeploymentConfig:
    """
    Custom CloudFormation resource which creates a git commit action to set deployment configuration for ecs project.
    """
    def __init__(
            self,
            stack: core.Stack,
            prefix: str,
            code_repository: Repository,
            task_definition: str,
            app_spec: str
    ) -> None:
        """
        Constructor.

        :param stack: A CloudFormation stack to which add this resource.
        :param prefix: Prefix for resource names.
        :param code_repository: A codecommit git repository to push configuration files for ecs deployment.
        :param task_definition: A document which describes how ecs deployment should behave.
        :param app_spec: A document which describes how ecs deployment should behave.
        """
        self.__stack = stack
        self.__prefix = prefix
        self.__code_repository = code_repository
        self.__task_definition = task_definition
        self.__app_spec = app_spec

    def get_resource(self):
        """
        Creates a custom resource to manage an ecs deployment configuration.

        :param scope: A scope in which this resource should be created.

        :return: Custom resource to manage an ecs deployment configuration.
        """
        return AwsCustomResource(
            self.__stack,
            self.__prefix + "CustomDeploymentConfigResource",
            on_create=self.__on_create(),
            on_update=self.__on_update(),
            on_delete=self.__on_delete(),
            role=self.__role()
        )

    def __role(self) -> Role:
        """
        A role for custom resource which manages ecs deployment configuration.

        :return: Custom resource's role.
        """
        return Role(
            self.__stack,
            self.__prefix + 'FargateDeploymentConfigRole',
            inline_policies={
                self.__prefix + 'FargateDeploymentConfigPolicy': PolicyDocument(
                    statements=[
                        PolicyStatement(
                            actions=[
                                "codecommit:CreateCommit",
                            ],
                            resources=[self.__code_repository.repository_arn],
                            effect=Effect.ALLOW
                        ),
                        PolicyStatement(
                            actions=[
                                "logs:CreateLogGroup",
                                "logs:CreateLogStream",
                                "logs:PutLogEvents"
                            ],
                            resources=['*'],
                            effect=Effect.ALLOW
                        )
                    ]
                )},
            assumed_by=ServicePrincipal('lambda.amazonaws.com')
        )

    @staticmethod
    def service_name() -> str:
        """
        Returns a service name that this custom resource manages.

        :return: Service name.
        """
        return 'CodeCommit'

    def __on_create(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_create" command.

        :return: A dictionary command.
        """
        return {
            "service": self.service_name(),
            "action": "createCommit",
            "parameters": {
                'branchName': 'master',
                'repositoryName': self.__code_repository.repository_name,
                'commitMessage': 'Initial appspec and taskdef files.',
                'putFiles': [
                    {
                        'filePath': 'taskdef.json',
                        'fileMode': 'NORMAL',
                        'fileContent': self.__task_definition,
                    }, {
                        'filePath': 'appspec.yaml',
                        'fileMode': 'NORMAL',
                        'fileContent': self.__app_spec,
                    }
                ]
            },
            "physical_resource_id": PhysicalResourceId.of(self.__prefix + 'CreateCommit')
        }

    def __on_update(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_update" command".

        :return: A dictionary command.
        """
        return None

    def __on_delete(self) -> Optional[Dict[Any, Any]]:
        """
        Creates an "on_delete" command".

        :return: A dictionary command.
        """
        return None



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup(
    name='aws_ci_cd_fargate',
    version='7.0.0',
    license='GNU GENERAL PUBLIC LICENSE Version 3',
    packages=find_packages(exclude=['venv', 'test']),
    description=(
        'AWS CDK package that helps deploying a fargate service with ci/cd.'
    ),
    long_description=README + '\n\n' + HISTORY,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        'aws_cdk.core>=1.27.0,<1.28.0',
        'aws_cdk.aws_iam>=1.27.0,<1.28.0',
        'aws_cdk.custom_resources>=1.27.0,<1.28.0',
        'aws_cdk.aws_s3>=1.27.0,<1.28.0',
        'aws_cdk.aws_certificatemanager>=1.27.0,<1.28.0',
        'aws_cdk.aws_elasticloadbalancingv2>=1.27.0,<1.28.0',
        'aws_cdk.aws_ec2>=1.27.0,<1.28.0',
        'aws_cdk.aws_logs>=1.27.0,<1.28.0',
        'aws_cdk.aws_ecs>=1.27.0,<1.28.0',
        'aws_cdk.aws_applicationautoscaling>=1.27.0,<1.28.0',
        'aws_cdk.aws_codedeploy>=1.27.0,<1.28.0',
        'aws_cdk.aws_codecommit>=1.27.0,<1.28.0',
        'aws_cdk.aws_codepipeline>=1.27.0,<1.28.0',
        'aws_cdk.aws_codepipeline_actions>=1.27.0,<1.28.0',
        'aws_cdk.aws_ecr>=1.27.0,<1.28.0',
        'aws_cdk.aws_codebuild>=1.27.0,<1.28.0',
        'aws-empty-bucket>=2.0.1,<3.0.0',
        'aws-vpc>=2.0.0,<3.0.0',
        'aws_ecs_service>=1.0.7,<2.0.0',
        'aws-ecs-cluster>=1.0.1,<2.0.0',
        'aws-empty-ecr-repository>=1.0.2,<2.0.0'
    ],
    author='Deividas Tamkus, Laimonas Sutkus',
    author_email='dtamkus@gmail.com (deividas@idenfy.com), laimonas.sutkus@gmail.com (laimonas@idenfy.com)',
    keywords='AWS CDK Fargate ECS',
    url='https://github.com/idenfy/AwsFargateCdk.git',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
)



from aws_cdk import aws_codepipeline, aws_codepipeline_actions, aws_codecommit, aws_codedeploy, aws_elasticloadbalancingv2, aws_ecs, aws_ecr
from aws_cdk.aws_ecs import CfnService
from aws_cdk.aws_s3 import IBucket
from aws_cdk.core import Stack
from aws_ci_cd_fargate.source.custom.deployment_config import DeploymentConfig
from aws_ci_cd_fargate.source.custom.deployment_group import DeploymentGroup


class PipelineEcrToEcs:
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            artifacts_bucket: IBucket,
            source_repository: aws_codecommit.Repository,
            ecr_repository: aws_ecr.Repository,
            task_def: str,
            app_spec: str,
            main_listener: aws_elasticloadbalancingv2.CfnListener,
            deployments_listener: aws_elasticloadbalancingv2.CfnListener,
            ecs_cluster: aws_ecs.Cluster,
            ecs_service: CfnService,
            production_target_group,
            deployment_target_group
    ):
        self.application = aws_codedeploy.EcsApplication(
            scope, prefix + 'FargateCodeDeployApplication',
            application_name=prefix + 'FargateCodeDeployApplication',
        )

        self.deployment_group_custom = DeploymentGroup(
            stack=scope,
            prefix=prefix,
            code_repository=source_repository,
            task_definition=task_def,
            app_spec=app_spec,
            ecs_application=self.application,
            main_listener=main_listener,
            deployments_listener=deployments_listener,
            ecs_cluster=ecs_cluster,
            production_target_group=production_target_group,
            deployment_target_group=deployment_target_group
        ).get_resource()

        self.deployment_group_custom.node.add_dependency(ecs_service)
        self.deployment_group_custom.node.add_dependency(ecs_cluster)

        self.deployment_group = aws_codedeploy.EcsDeploymentGroup.from_ecs_deployment_group_attributes(
            scope, prefix + 'FargateDeploymentGroup',
            application=self.application,
            deployment_group_name=prefix + 'FargateDeploymentGroup',
        )

        self.deployment_group.node.add_dependency(self.deployment_group_custom)

        self.deployment_config_repository = aws_codecommit.Repository(
            scope, prefix + 'FargateDeploymentConfigRepository',
            description='Repository containing appspec and taskdef files for ecs code-deploy blue/green deployments.',
            repository_name=prefix.lower() + '-deployment-config'
        )

        self.commit_custom = DeploymentConfig(
            stack=scope,
            prefix=prefix,
            code_repository=self.deployment_config_repository,
            task_definition=task_def,
            app_spec=app_spec
        ).get_resource()

        self.ecr_repository_output_artifact = aws_codepipeline.Artifact('EcsImage')
        self.config_output_artifact = aws_codepipeline.Artifact('EcsConfig')

        self.ecr_to_ecs_pipeline = aws_codepipeline.Pipeline(
            scope,
            prefix + 'FargateEcrToEcsPipeline',
            artifact_bucket=artifacts_bucket,
            pipeline_name=prefix + 'FargateEcrToEcsPipeline',
            stages=[
                aws_codepipeline.StageProps(
                    stage_name='SourceStage',
                    actions=[
                        aws_codepipeline_actions.EcrSourceAction(
                            action_name='SourceEcrAction',
                            output=self.ecr_repository_output_artifact,
                            repository=ecr_repository,
                            run_order=1,
                        ),
                        aws_codepipeline_actions.CodeCommitSourceAction(
                            action_name='SourceCodeCommitAction',
                            output=self.config_output_artifact,
                            repository=self.deployment_config_repository,
                            branch='master',
                            run_order=1,
                        )
                    ]
                ),
                aws_codepipeline.StageProps(
                    stage_name='DeployStage',
                    actions=[
                        aws_codepipeline_actions.CodeDeployEcsDeployAction(
                            action_name='DeployAction',
                            deployment_group=self.deployment_group,
                            app_spec_template_input=self.config_output_artifact,
                            task_definition_template_input=self.config_output_artifact,
                            container_image_inputs=[
                                aws_codepipeline_actions.CodeDeployEcsContainerImageInput(
                                    input=self.ecr_repository_output_artifact,
                                    task_definition_placeholder='IMAGE1_NAME'
                                )
                            ],
                            run_order=1
                        )
                    ]
                )
            ]
        )

        self.ecr_to_ecs_pipeline.node.add_dependency(self.commit_custom)



from typing import List, Dict, Any
from aws_cdk import aws_ec2


class EcsParams:
    """
    Parameters class which specifies deployed container and ecs parameters such as name, port, etc.
    """
    def __init__(
            self,
            container_name: str,
            container_cpu: int,
            container_ram: int,
            container_environment: Dict[str, Any],
            ecs_security_groups: List[aws_ec2.SecurityGroup],
            ecs_subnets: List[aws_ec2.Subnet],
            cpu_threshold: int = 50
    ) -> None:
        """
        Constructor.

        :param container_name: The name that will be given to a newly deployed container.
        :param container_cpu: Cpu units for the deployed container. 1 CPU = 1024 Cpu units.
        :param container_ram: Memory for the deployed container. 1 GB Ram = 1024 units.
        :param container_environment: Environment that will be passed to a running container.
        :param ecs_security_groups: Security groups for ecs service in which containers are placed.
        :param ecs_subnets: Subnets to which new containers will be deployed.
        :param cpu_threshold: Measured in percent. 0-100 percent. This value indicates when an autoscaling should
        kick in. If an average containers' cpu utilization is below this threshold, the amount of servers should be
        decreased. On the contrary, if an average containers' cpu utilization is above this threshold, the amount
        of servers will be increased.

        :return: No return.
        """
        self.container_name = container_name
        self.container_cpu = container_cpu
        self.container_ram = container_ram
        self.container_environment = container_environment
        self.ecs_security_groups = ecs_security_groups
        self.ecs_subnets = ecs_subnets
        self.cpu_threshold = cpu_threshold
